# -*- coding: utf-8 -*-
'''
Created on 30.06.2010

@author: adm
'''
import logging
import copy
from impact import Impact, ImpactObjectCheck, ImpactObjectPosition, ImpactGroup, ImpactMode
from objectdiscount import CheckObject, PositionObject, Kit, LegacyObject, SimpleObject
from roster import Roster
from campaign import DiscountCampaign
from artixds.kitlogic.kitlogic import KitLogic
from artixds.kitlogic.kitobject import KitObject

log = logging.getLogger("Discount")

class Discount(object):
    """
    Discount
    
    """
    
    coupons = []
    counters = []
    def __init__(self, campaign, objects=None, conditions=None, discountRate=None, id=None, name=None, cashText = None,
                 clientText = None, minPriceIgnored = False, active=True, coupons=None, counters=None):
        """
        campaign - parent DicountCampaign
        objects[] - list of Object
        conditions[] - list of Condition
        discountRate - DiscountRate
        """
        self.campaign = campaign
        self.objects = objects
        self.conditions = conditions
        if(self.conditions == None):
            self.conditions = []
        self.discountRate = discountRate
        self.id = id
        self.name = name
        self.cashText = cashText
        self.clientText = clientText
        self.minPriceIgnored = minPriceIgnored
        self.active = active
        
        self.coupons = coupons
        if(self.coupons == None):
            self.coupons = []
            
        self.counters = counters
        if(self.counters == None):
            self.counters = []

    def isActive(self):
        return self.active if hasattr(self, 'active') else True

    def isValidKitObject(self, checkObj):
        if isinstance(checkObj, Roster):
            for obj in filter(lambda x: isinstance(x, Kit), self.objects):
                res = True if checkObj.findKit(obj) else False
                checkObj.currentKit = obj
                if(res):
                    return True
        return False 
    
    def presenceObjects(self):
        checkObj, posObj, kitObj, legObj, kitNewObj, simpleObj = False, False, False, False, False, False
        for obj in self.objects:
            if not checkObj and isinstance(obj, CheckObject):
                checkObj = True
            if not posObj and isinstance(obj, PositionObject):
                posObj = True
            if not kitNewObj and isinstance(obj, KitObject): 
                kitNewObj = True
            if not kitObj and isinstance(obj, Kit): 
                kitObj = True
            if not legObj and isinstance(obj, LegacyObject): 
                legObj = True
            if not simpleObj and isinstance(obj, SimpleObject): 
                simpleObj = True
            if checkObj and posObj and kitObj and legObj:
                break
        return checkObj, posObj, kitObj, legObj, kitNewObj, simpleObj

    def isConditionsTrue(self, obj, check, discountContext):
        """
        Check all conditions.
        
        obj - current Object
        check - Check
        
        Возвращает 2 значения:
        1. True, если все условия выполняются, иначе False
        2. True, если какое то условие не зависящее от позиции не выполнено и дальнейший расчет бесполезен, иначе False
        """

        # Проверка выполнения условия.
        def validate(conditionsList):
            for condit in conditionsList:
                condition = condit
                condition.checkWrapper = check
                object = obj.data if hasattr(obj, "data") else obj.currentKit
                # если хотя бы одно условие ложно, то завершаем расчет скидок для этой акции
                if not condition.validate(check.checkDict, object, discountContext) :
                    return False
            return True

        # расчитываем условия не зависящие от позиций
        if not self.notPosConditionTrue:
            # если хотя бы одно условие ложно, то завершаем расчет скидок для этой акции
            if not validate(discountContext.posIndependent):
                return False, True
            self.notPosConditionTrue = True 

        # расчитываем условия зависящие от позиции
        if not validate(discountContext.posDependent):
            return False, False,

        return True, False

    def injectPosition(self, check, qty, currentNumber, data, partOfDisKit):
        """
        Create new position and inject it to check
        """
        log.debug("inject position number %s", currentNumber)
        position = data
        position["posnum"] = currentNumber
        position["sumn"] = position["price"] * qty
        position["sumb"] = position["sumb"] / position["cquant"] * qty
        position["cquant"] = qty
        # TODO: add recalculation "sum"
        position["partOfDisKit"] = partOfDisKit
        check.addPosition(position)


    def getGoods(self, obj, check, code):
        res = filter(lambda obj:isinstance(obj, PositionObject) and obj.data["code"] == code, check.getObjects())
        if len(res) < 0:
            raise Exception("Couldn't find goods with code %s in check" % code)
        goods = res[0]
        return  goods

    def getKitImpact(self, *args):
        '''
        kit - Kit
        check - checkWrapper
        '''
        check = args[0]
        obj = args[1]

        check.isModified = True
        self.discountRate.checkWrapper = check
        impactGroup = ImpactGroup()
        kit = obj.currentKit
        entry = obj.findKit(kit)
        applytimes = 0
        while entry and kit.permissionApplyKit(applytimes):
            applytimes += 1
            obj.subtractKitEntry(entry)
            entry = obj.findKit(kit)
        obj.applyRate(self.discountRate, self.campaign, self)
        pureImpacts = obj.pureImpacts()
        structureImpacts = obj.structureImpacts()
        impactGroup.addImpacts(structureImpacts)
        impactGroup.addImpacts(pureImpacts)
        for impact in structureImpacts:
            check.branchPosition(impact.impactObject.positionNumber, impact.branchDetails)
        for posnum in obj.partsOfKit():
            check.markAsPartOfKit(posnum)
        for impact in pureImpacts:
            check.positionsImpact.add(impact.impactObject.positionNumber)

        return impactGroup

    def getKitNewImpact(self, *args):
        '''
        Расчет скидок на новый набор
        '''
        check = args[0]

        check.isModified = True
        impactGroup = ImpactGroup()
        self.discountRate.checkWrapper = check
        kit = None
        
        for obj in filter(lambda x: isinstance(x, KitObject), self.objects):
            kit = obj
            break
        kitLogic = KitLogic(check, kit, self.discountRate, self.campaign, self)
        if not kitLogic.findKit():
            return impactGroup

        impactGroup.addImpacts(kitLogic.uniteImpacts)
        impactGroup.addImpacts(kitLogic.structureImpacts)
        impactGroup.addImpacts(kitLogic.pureImpacts)

        for impact in kitLogic.uniteImpacts:
            check.unitePosition(impact.impactObject.positionNumber, impact.qty, impact.uniteDetails)
        for impact in kitLogic.structureImpacts:
            check.branchPosition(impact.impactObject.positionNumber, impact.branchDetails)
        for posnum in kitLogic.partsOfKit:
            check.markAsPartOfKit(posnum)
        for impact in kitLogic.pureImpacts:
            check.positionsImpact.add(impact.impactObject.positionNumber)
        return impactGroup

    def getCheckImpact(self, *args):
        check = args[0]
        obj = args[1]
        discountContext = args[2]
        checkDict = check.checkDict 
        self.discountRate.checkWrapper = check

        check.isModified = True
        impactGroup = ImpactGroup()
        impact = Impact()
        impact.campaign = self.campaign
        impact.discount = self
        impact.innerObject = obj
        impact.impactObject = ImpactObjectCheck()
        card = discountContext.cards[0] if discountContext.cards else None
        impact.absoluteValue = self.discountRate.getAbsoluteValue(obj, checkDict, card)
        impact.cards = discountContext.cards
        impact.coupons = discountContext.coupons
        impact.counters = discountContext.counters

        if impact.discountMode == ImpactMode.AUTOMATIC:
            if discountContext.coupons:
                impact.discountMode = ImpactMode.BYCOUPON
            elif discountContext.cards:
                impact.discountMode = ImpactMode.BYCARD        
        
        impactGroup.addImpact(impact)
        return impactGroup

    def getPosImpact(self, *args):
        check = args[0]
        obj = args[1]
        discountContext = args[2]  
        checkDict = check.checkDict 
        self.discountRate.checkWrapper = check

        check.isModified = True
        impactGroup = ImpactGroup()
        impact = Impact()
        impact.campaign = self.campaign
        impact.discount = self
        impact.innerObject = obj 
        impact.impactObject = ImpactObjectPosition(positionNumber=obj.data["posnum"], code=obj.data["code"])
        impact.qty = obj.data["cquant"]
        card = discountContext.cards[0] if discountContext.cards else None 
        impact.absoluteValue = self.discountRate.getAbsoluteValue(obj, checkDict, card)
        impact.cards = discountContext.cards
        impact.coupons = discountContext.coupons

        if impact.discountMode == ImpactMode.AUTOMATIC:
            if discountContext.coupons:
                impact.discountMode = ImpactMode.BYCOUPON
            elif discountContext.cards:
                impact.discountMode = ImpactMode.BYCARD
                
        impactGroup.addImpact(impact)
        check.positionsImpact.add(obj.data["posnum"])
        return impactGroup


    def getLegacyImpact(self, *args):

        from legacydiscount import main as legacySystem
        from branchdetail import BranchDetail

        check = args[0]
        discountContext = args[2]
        self.discountRate.checkWrapper = check
        check.isModified = True
        impactGroup = ImpactGroup()
        
        import artix_core
        b,d = legacySystem.calculateImpact(check.check, artix_core.config)
        structureImpacts = []
        pureImpacts = []

        for i in b:
            branchDetails = []
            for sImp in b[i]:
                detail = BranchDetail(sImp['posnum'], sImp['qty'], sImp['sum'], sImp['sum'], False)
                branchDetails.append(detail)
            structureImpacts.append(Impact.branchImpact(ImpactObjectPosition(int(i), None), branchDetails))

        for impact in structureImpacts:
            check.branchPosition(impact.impactObject.positionNumber, impact.branchDetails)

        for dImp in d:
            impact = Impact()
            impact.cards = []
            impact.campaign = DiscountCampaign(id = dImp.rate.scheme.code, name = dImp.rate.scheme.name)

            code = 0
            if hasattr(dImp.innerObject, 'code'):
                code = dImp.innerObject.code
            impact.impactObject = ImpactObjectPosition(positionNumber=dImp.innerObject.posnum, code=code)
            impact.qty = dImp.innerObject.cquant
            impact.absoluteValue = float(dImp.absoluteValue)

            if dImp.card:
                if isinstance(dImp.card.dump(), dict):
                    impact.cards.append(dImp.card.dump())
                else:
                    impact.cards.append({'codeinternal' : 1})

            if dImp.discountMode == "bydisckit":
                impact.discountMode = ImpactMode.BYDISCKIT
                check.markAsPartOfKit(dImp.innerObject.posnum)
            elif dImp.discountMode == "bycard":
                impact.discountMode = ImpactMode.BYCARD
            elif dImp.discountMode == "automatic":
                impact.discountMode = ImpactMode.AUTOMATIC

            if impact.discountMode == ImpactMode.AUTOMATIC and discountContext.cards:
                impact.discountMode = ImpactMode.BYCARD
            pureImpacts.append(impact)

            self.discountRate.type = dImp.rate.type
            self.discountRate.value = float(dImp.rate.value)
            self.discountRate.items = []
            self.discountRate.checkWrapper = None

            impact.discount = Discount(self)
            impact.discount.id = dImp.rate.disccode
            impact.discount.name = dImp.rate.name
            impact.discount.discountRate = copy.deepcopy(self.discountRate)
            impact.discount.cashText = copy.deepcopy(self.cashText)
            impact.discount.clientText = copy.deepcopy(self.clientText)
            impact.discount.minPriceIgnored = copy.deepcopy(self.minPriceIgnored)
            impact.discount.active = copy.deepcopy(self.active)
            check.markPositionByDiscount(dImp.innerObject.posnum)

        impactGroup.addImpacts(structureImpacts)
        impactGroup.addImpacts(pureImpacts)
        return impactGroup
    
    def getSimpleImpact(self, *args):

        from simplediscount import main as simpleSystem
        from branchdetail import BranchDetail

        check = args[0]
        discountContext = args[2]
        self.discountRate.checkWrapper = check
        check.isModified = True
        impactGroup = ImpactGroup()
        
        import artix_core
        b,d = simpleSystem.calculateImpact(check.check, artix_core.config)
        structureImpacts = []
        pureImpacts = []

        for i in b:
            branchDetails = []
            for sImp in b[i]:
                detail = BranchDetail(sImp['posnum'], sImp['qty'], sImp['sum'], sImp['sum'], False)
                branchDetails.append(detail)
            structureImpacts.append(Impact.branchImpact(ImpactObjectPosition(int(i), None), branchDetails))

        for impact in structureImpacts:
            check.branchPosition(impact.impactObject.positionNumber, impact.branchDetails)

        for dImp in d:
            impact = Impact()
            impact.cards = []
            impact.campaign = DiscountCampaign(id = dImp.rate.scheme.code, name = dImp.rate.scheme.name)

            code = 0
            if hasattr(dImp.innerObject, 'code'):
                code = dImp.innerObject.code
            impact.impactObject = ImpactObjectPosition(positionNumber=dImp.innerObject.posnum, code=code)
            impact.qty = dImp.innerObject.cquant
            impact.absoluteValue = float(dImp.absoluteValue)

            if dImp.card:
                if isinstance(dImp.card.dump(), dict):
                    impact.cards.append(dImp.card.dump())
                else:
                    impact.cards.append({'codeinternal' : 1})

            if dImp.discountMode == "bydisckit":
                impact.discountMode = ImpactMode.BYDISCKIT
                check.markAsPartOfKit(dImp.innerObject.posnum)
            elif dImp.discountMode == "bycard":
                impact.discountMode = ImpactMode.BYCARD
            elif dImp.discountMode == "automatic":
                impact.discountMode = ImpactMode.AUTOMATIC

            if impact.discountMode == ImpactMode.AUTOMATIC and discountContext.cards:
                impact.discountMode = ImpactMode.BYCARD
            pureImpacts.append(impact)

            self.discountRate.type = dImp.rate.type
            self.discountRate.value = float(dImp.rate.value)
            self.discountRate.items = []
            self.discountRate.checkWrapper = None

            impact.discount = Discount(self)
            impact.discount.id = dImp.rate.disccode
            impact.discount.name = dImp.rate.name
            impact.discount.discountRate = copy.deepcopy(self.discountRate)
            impact.discount.cashText = copy.deepcopy(self.cashText)
            impact.discount.clientText = copy.deepcopy(self.clientText)
            impact.discount.minPriceIgnored = copy.deepcopy(self.minPriceIgnored)
            impact.discount.active = copy.deepcopy(self.active)
            check.markPositionByDiscount(dImp.innerObject.posnum)

        impactGroup.addImpacts(structureImpacts)
        impactGroup.addImpacts(pureImpacts)
        return impactGroup