# -*- coding: utf-8 -*-
'''
Created on 18.06.2010

@author: adm
'''

#from configuration import logging
import logging
import enum
from ratefunction import RateFunction
log = logging.getLogger("DiscountRate")

class DiscountRateType(enum.Enum):
    PERCENT = enum.Item()
    SUM = enum.Item()
#    CALCULATE = enum.Item()
#    KIT = enum.Item()
    INDEX = enum.Item()

def createDiscountRate(type, value):
    return {DiscountRateType.PERCENT : PercentDiscountRate,
            DiscountRateType.SUM     : SumDiscountRate,
            DiscountRateType.INDEX   : IndexDiscountRate}.get(type, lambda x : None)(value)

class DiscountRate(object):
    def __init__(self, type=None):
        self.type = type   
        self.checkWrapper = None		
       # self.card = None 		
        
    def getAbsoluteValue(self, obj, check, card=None):
        """
        check - current checkWrapper
        # TODO: what's kind of obj? Different  for different DiscountRate?
        obj - ????
        """
        return 0
        
    def getRateType(self):
        return self.type
        
class PercentDiscountRate(DiscountRate):
    def __init__(self, value=None):
        DiscountRate.__init__(self, DiscountRateType.PERCENT)
        self.value = value

    def getAbsoluteValue(self, obj, check, card=None ):
        return obj.getPercentOfFinalSum(self.value)
    def __repr__(self):
        return "%s%%" % self.value

class SumDiscountRate(DiscountRate):
    def __init__(self, value=None):
        DiscountRate.__init__(self, DiscountRateType.SUM)
        self.value = value

    def getAbsoluteValue(self, obj, check, card=None):
        return obj.getAbsoluteDiscountSum(self.value)

class IndexDiscountRate(DiscountRate):
    def __init__(self, value=None):
        DiscountRate.__init__(self, DiscountRateType.INDEX)
        self.value = value
    def getAbsoluteValue(self, obj, check, card=None):
        return obj.getIndexDiscountSum(int(self.value))

class CalculateDiscountRate(DiscountRate):
    def __init__(self, value=None, type=DiscountRateType.PERCENT):
        DiscountRate.__init__(self, type)  
        self.value = value

    def eval(self, rf, object, check, card=None):
        return eval(self.value)
    
    def getAbsoluteValue(self, obj, check, card=None):
        try:
            log.info('calculating rate using expression: %s' % self.value)
            log.info('Curent card info: %s' % card)
            rf = RateFunction(check, obj, card, self.checkWrapper)
            val = self.eval(rf, obj.data, check, card)
        except Exception, e:
            log.error("error in eval call: %s" % e)
            log.exception(e)
            val = 0
        log.info('calculated value is : %s' % val)
        log.info('type is: %s' % self.type)
        return createDiscountRate(self.type, val).getAbsoluteValue(obj, check)
	
class KitDiscountRateItem(object):
    kitItem = None
    discountRate = None

    def __init__(self, kitItem = None, discountRate = None):
        self.kitItem = kitItem
        self.discountRate = discountRate

    def __repr__(self):
        return "KitDiscountRateItem(kitItem=%s,discountRate=%s)" % (self.kitItem, self.discountRate)

class KitDiscountRate(DiscountRate):
    items = []
    
    def __init__(self, items = []):
        if items:
            self.items = items
        else:
            self.items = []

    def __repr__(self):
        return "KitDiscountRate(items=%s)" % self.items

    def addItem(self, kitItem, discountRate):
        self.items.append(KitDiscountRateItem(kitItem, discountRate))

    def __iter__(self):
        for item in self.items:
            yield item

    def discountRateForItem(self, kitItem):
        for item in self:
            if item.kitItem == kitItem:
                return item.discountRate
        return None

    def getAbsoluteValues(self, kitEntryItem, card=None):
        
        return []
