# -*- coding: utf-8 -*-
'''
Created on 16.06.2010

http://gist.github.com/401547
Python Enum от Димы Смолина 14.05.2010 


Дима пишет: "В кои-то веки сделал Python реализацию Enum,
 которая мне нравится :) Опыт использования показал, что пользоваться 
 очень удобно :) Идея трюка с определением порядка атрибутов класса 
 сперта из Django. Мы используем строковые константы, но если кому-то 
 нужна числовая автонумерация - код очень легко дополнить"
import enum

class TrackAction(enum.Enum):
    PLAY  = enum.Item()
    STOP  = enum.Item(choice='Stop! Stop!')
    PAUSE = enum.Item(value='P')


print TrackAction.PLAY
# play

print TrackAction.STOP
# stop

print TrackAction.PAUSE
# P

print TrackAction.values
# ['play', 'stop', 'P']

print TrackAction.keys
# ['PLAY', 'STOP', 'PAUSE']

print TrackAction.choices
# (('play', 'Play'), ('stop', 'Stop! Stop!'), ('P', 'Pause'))

print r'^track/(?P<id>\d+)/(?P<event>'+TrackAction.url_re()+')/$'
# ^track/(?P<id>\d+)/(?P<event>play|stop|P)/$
# (can be used in urlconf)
http://gist.github.com/401547
'''

class Unset(object):
    pass

class Item(object):
    _cnt = 0
    
    def __init__(self, value=Unset, choice=Unset):
        self.key = None
        self.value = value
        self.choice = choice
        self.num = Item._cnt
        Item._cnt += 1

class EnumMeta(type):
    def __new__(cls, name, bases, attrs):
        items = []
        for k, v in attrs.items():
            if isinstance(v, Item):
                v.key = k
                if v.value is Unset:
                    v.value = k.lower()
                if v.choice is Unset:
                    v.choice = k.capitalize()
                items.append(v)
                attrs[k] = v.value
        items.sort(key=lambda item: item.num)
        attrs['keys'] = [item.key for item in items]
        attrs['values'] = [item.value for item in items]
        if not 'choices' in attrs:
            attrs['choices'] = tuple((item.value, item.choice) for item in items)
        return super(EnumMeta, cls).__new__(cls, name, bases, attrs)

class Enum(object):
    __metaclass__ = EnumMeta
    
    @classmethod
    def url_re(cls):
        return '|'.join(str(v) for v in cls.values)
