# -*- coding: utf-8 -*-
'''
Created on 16.06.2010

@author: adm
'''

import copy
from datetime import datetime
from artixds.domain import  DiscountCampaign, DiscountCampaignGroup, Coupon, Counter, CounterChanged
from artixds.checkwrapper  import CheckWrapper
from artixds.domain import aggregateImpact, ImpactGroup, InteractionType, ImpactType
from artixds.conditionfunctionplugins.conditionplugin import conditionMemory
from artixds.ratefunctionplugins.rateplugin import rateMemory
import jsonpickle
import yaml
import re
import time
import logging
from artixds.conditionfunction import ConditionFunction
from artixds.objectdiscount import CheckObject

#logging.config.fileConfig("logging.properties")
logging.basicConfig(level=logging.INFO,format='[%(levelname)s]-%(name)s: %(message)s \nFile "%(pathname)s", line %(lineno)d')

log = logging.getLogger("Subsystem")

def timer(func):
    '''
        Декоратор для измерения времени выполнения функции func
    '''
    def measure(*args, **kwargs):
        timeStart = time.time()
        res = func(*args, **kwargs)
        log.info('Run-time function %s: %f: ' % ( func.__name__,  time.time() - timeStart))
        return res
    return measure

def getStrCheckPositions(check):
    if log.isEnabledFor(logging.DEBUG):
        log.debug("print check positions")
        log.debug("object id = %d" % id(check))
        for goodsItem in check.cwGoodsList:
            partOfDisKit = "partOfKit" in goodsItem and  goodsItem["partOfKit"] == True
            log.debug("posnum:%2d,code:%s,qty:%3s,%s" % (goodsItem["posnum"], goodsItem["code"], \
                goodsItem["cquant"], partOfDisKit))

def filterKeys(d):
    lowerD = {}
    for k in d.keys():
        if d[k] != None and d[k] != []:
            lowerD[k.lower()] = d[k]
    return lowerD

checkNumerator = 0
firstAdd = []
def getCurrentCheck(check, campaignGroup):
    """
    Return check or its copy
    """
    if campaignGroup.interactionType == InteractionType.ALL:
        global checkNumerator
        if campaignGroup.id in firstAdd and check.isModified:
            tmp = checkNumerator
            checkNumerator = tmp + 1
            check.id = checkNumerator
        else:
            firstAdd.append(campaignGroup.id)
        check.isModified = False
        return check
    return CheckWrapper().copy(check)


class BaseImpacts(object):
    """
    Create impacts at the root campaign hierarchy

    """
    def __init__(self, check, rootCampaignGroup, discountSystemNumber):
        self.check = check
        self.rootCampaignGroup = rootCampaignGroup
        self.discountSystemNumber = discountSystemNumber

    @timer
    def getImpacts(self):
        return self.getGroupImpacts(self.check, self.rootCampaignGroup)

    def  getGroupImpacts(self, check, campaignGroup):
        """
        Get impact for CampaignGroup

        """
        allImpacts = []
        log.debug("work with campaign group:%s", campaignGroup)
        check.newGroup()

        """
        objects - objects of discountCampaigns[] + objects of childGroups[]
        """
        objects = campaignGroup.discountCampaigns + campaignGroup.childGroups
        interaction = campaignGroup.interactionType
        # InteractionType.LAST сводится к InteractionType.FIRST сменой порядка применения акций(условий) на обратный
        if interaction == InteractionType.LAST:
            objects.reverse()

        for obj in objects:
            
            if isinstance(obj, DiscountCampaign):
                currentCheck = getCurrentCheck(check,campaignGroup)
                getStrCheckPositions(currentCheck)
                campaign = DiscountCampaign()
                campaign = obj
                log.debug("campaign=%s", campaign.name)
                if(campaign.isValidDate(check.getDataTimeBeg()) and campaign.isActive()):
                    activatedCampaigns = []
                    if('activatedcampaigns' in check.checkDict):
                        activatedCampaigns = check.checkDict["activatedcampaigns"]
                    if not campaign.manualActivate or (str(self.discountSystemNumber) in activatedCampaigns and campaign.id in activatedCampaigns[str(self.discountSystemNumber)]):
                        cashTags = []
                        if('context' in check.checkDict and 'shoplabels' in check.checkDict["context"]):
                            cashTags = check.checkDict["context"]["shoplabels"].split(",")
                        if campaign.getCheckedTags() is None or campaign.isValidTag(cashTags):
                            log.debug("campaign date is valid and campaign is active and shop tags is valid")
                            impact = DiscountCampaignImpact(campaign, currentCheck)
                            if impact:
                                impactList = impact.getImpacts()
                                allImpacts.append(impactList)
                                if impactList.getImpacts() and interaction in [InteractionType.FIRST, InteractionType.LAST]:
                                    break
                        else:
                            log.debug("Shop tags don't valid")
                    else:
                        log.debug("Campaign for manual activation is not activated")
                else:
                    log.debug("campaign date don't valid")

            elif isinstance(obj, DiscountCampaignGroup):
                currentCheck = getCurrentCheck(check, campaignGroup)
                getStrCheckPositions(currentCheck)
                impact = self.getGroupImpacts(currentCheck, obj)
                if impact and impact.check:
                    getStrCheckPositions(impact.check)
                    allImpacts.append(impact)
                    if impact.getImpacts() and interaction in [InteractionType.FIRST, InteractionType.LAST]:
                        break

        if not campaignGroup.childGroups:
            log.debug("current group don't have child group")

        # отметим позиции, на которые применено воздействие
        impacts = aggregateImpact(allImpacts, campaignGroup.interactionType)
        if impacts:
            for impact in filter(lambda x:x.type == ImpactType.PURE,impacts.getImpacts()):
                if impact.impactObject.positionNumber != 0:
                    check.markPositionByDiscount(impact.impactObject.positionNumber)

        return impacts

class DiscountCampaignImpact(object):
    """
    Create impacts at the DiscountCampaign

    """
    def __init__(self, campaign, check):
        self.campaign = campaign
        self.check = check

    @timer
    def getImpacts(self):
        """
        Get impacts for discount campaign

        """
        campaign = self.campaign
        interaction = campaign.interactionType
        allImpacts = []
        self.check.newGroup()

        if interaction == InteractionType.LAST and not campaign.isDiscountListReverse():
            campaign.setDiscountListInReverse()
        for discount in campaign.discounts:
            if discount.isActive():
                currentCheck = getCurrentCheck(self.check, campaign)
                getStrCheckPositions(currentCheck)
                impact = DiscountImpacts(discount)
                impactGroup = impact.getImpact(currentCheck)
                getStrCheckPositions(currentCheck)
                allImpacts.append(impactGroup)
                if impactGroup.getImpacts() and interaction in [InteractionType.FIRST, InteractionType.LAST]:
                    break
        return aggregateImpact(allImpacts, interaction)

class DiscountContext(object):
    """
    Контекст скидки. При принятии решения о применения каждой новой скидки
    создаётся контекст скидки, который существует до момента образования
    воздействия
    """
    cards = []
    coupons = []

    def __init__(self):
        self.cards = []
        self.coupons = []
        self.counters = []
        self.posDependent = []
        self.posIndependent = []
    def __repr__(self):
        return repr(self.cards)

class DiscountImpacts(object):
    """
    Create impacts at the Discount

    """

    def __init__(self, discount):
        """
        discount - Discount
        posDependent - list of conditions, depending on the position
        posIndependent = list of conditions, independing on the position
        """
        self.discount = discount


    def selectImpacts(self, obj, check, discount, discountContext, getImpactFunction):
        log.debug("work with object=%s", obj)
        specific, total  = discount.isConditionsTrue(obj, check, discountContext)
        if total:
            return None, True
        if specific:
            log.debug("all conditions true")
            log.debug("discountContext = %s" % discountContext)
            impact = getImpactFunction(check, obj, discountContext)
            return impact.getImpacts(), total
        else:
            log.debug("not all conditions true")
            return None, False

    def getImpact(self, check):
        """
        Get impacts for discount
        return single ImpactGroup
        """
        discount = self.discount
        discount.notPosConditionTrue = False
        log.debug("discount=%s", discount)
        impactGroup = ImpactGroup()
        cardForMacro = None
        checkObjExist, posObjExist, kitObjExist, legObjExist, kitNewObjExist, simpleObjExist = discount.presenceObjects()
        discountContext = DiscountContext()

        # формируем списки условий зависящих и не зависящих от позиций
        contentPosConditions = ConditionFunction().posConditionList
        for conditItem in discount.conditions:
            posDependentCondition = filter( lambda x: x in conditItem.condition, contentPosConditions)
            if (posDependentCondition):
                discountContext.posDependent.append(conditItem)
            else:
                discountContext.posIndependent.append(conditItem)

        if (posObjExist):
            log.debug(u"Расчет условий для позиций")
            for obj in check.getGoodsList():
                imp, res = self.selectImpacts(obj, check, discount, discountContext, discount.getPosImpact)
                if res:
                    break
                if imp:
                    impactGroup.addImpacts(imp)

        if(checkObjExist):
            log.debug(u"Расчет условий для чека")
            obj = check.getCheckObject()
            imp, res = self.selectImpacts(obj, check, discount, discountContext, discount.getCheckImpact)
            impactGroup.addImpacts(imp) if imp else False

        obj = check.createKit()
        if kitObjExist and discount.isValidKitObject(obj):
            log.debug(u"Расчет условий для наборов")
            imp, res = self.selectImpacts(obj, check, discount, discountContext, discount.getKitImpact)
            if imp:
                impactGroup.addImpacts(imp)

        if kitNewObjExist:
            log.debug(u"Расчет условий для наборов (новый алгоритм)")
            obj = check.getCheckObject()
            imp, res = self.selectImpacts(obj, check, discount, discountContext, discount.getKitNewImpact)
            if imp:
                impactGroup.addImpacts(imp)

        if legObjExist:
            log.debug(u"Расчет для старой дисконтной системы")
            obj = check.getCheckObject()
            imp, res = self.selectImpacts(obj, check, discount, discountContext, discount.getLegacyImpact)
            if imp:
                impactGroup.addImpacts(imp)

        if simpleObjExist:
            log.debug(u"Расчет для простой дисконтной системы")
            obj = check.getCheckObject()
            imp, res = self.selectImpacts(obj, check, discount, discountContext, discount.getSimpleImpact)
            if imp:
                impactGroup.addImpacts(imp)

        impactGroup.check = check

        # форматирование и сохранение сообщений для кассира, печати в чеке, купонов для печати, счетчиков
        if impactGroup.getImpacts():
            # добавляем сообщение, если скидка применена
            if log.isEnabledFor(logging.DEBUG):
                if discount.cashText:
                    log.debug(u'Сообщение на экран кассира: "%s"' % discount.cashText)
                    #log.debug(u'Сообщение на экран кассира: "%s"' % discount.cashText.encode('utf-8'))
                if discount.clientText:
                    log.debug(u'Сообщение для печати в чеке: "%s"' % discount.clientText)
                    #log.debug(u'Сообщение для печати в чеке: "%s"' % discount.clientText.encode('utf-8'))

            # запомним первую попавшуюся карту для того, чтобы использовать в макросах
            cardForMacro = discountContext.cards[0] if not cardForMacro and discountContext.cards else None

            if (discount.cashText and discount.cashText.strip()):
                impactGroup.screenMessage.append(self.__prepareMessage(discount.cashText, impactGroup.check, cardForMacro, impactGroup))
            if (discount.clientText and discount.clientText.strip()):
                impactGroup.receiptMessage.append(self.__prepareMessage(discount.clientText, impactGroup.check, cardForMacro, impactGroup))
            if (discount.coupons):
                for c in discount.coupons:
                    c.report =  self.__prepareMessage(c.report, impactGroup.check, cardForMacro, impactGroup)
                    c.campaignId = discount.campaign.id
                    c.campaignName = discount.campaign.name
                    impactGroup.coupons.append(c)

            # обработка расчетных функций для счетчиков дисконта
            if discount.counters and check.checkDict["cardrecords"]:
                for c in discount.counters:
                    log.debug(u"Счетчики из дисконта %s %s" % (c.counterId, c.rate))
                    c.rate.checkWrapper = check
                    value = c.rate.getAbsoluteValue(CheckObject(check.checkDict), check.checkDict)
                    if value > 0:
                        impactGroup.counters.append(CounterChanged(c.counterId, value, discount.campaign.id))
                    delattr(c.rate, "checkWrapper")

        return impactGroup

    def __prepareMessage(self, message, check = None, card = None, impact = None):
        log.debug("Prepare message \"%s\"", message)
        replaceBlocks = []
        for matchObject in re.finditer('\${(.+?)}', message):
            expression = matchObject.group(1)
            try:
                expression, format = expression.split(':', 1) if ':' in expression else [expression, None]
                value = self.__eval(expression, check.checkDict, card, impact)
                if str(type(value)) in ['<class \'PythonQt.QtCore.QDate\'>', '<class \'PythonQt.QtCore.QDateTime\'>', '<class \'PythonQt.QtCore.QTime\'>', \
                                        '<class \'check_test.QDateTime\'>']:
                    if not value.isNull():
                        value = value.toString(format) if format else value.toString()
                    else:
                        value = "---"
                else:
                    if value and not (value == "None" or value == ""):
                        value = ''.join(['{0:', format, '}']).format(value) if format else value
                    else:
                        value = "---"

                replaceBlocks.append((matchObject.group(0), value))
            except Exception, e:
                log.error("Prepare message. Error in eval call: '%s'" % expression)
                log.exception(e)
                replaceBlocks.append((matchObject.group(0), "---"))

        # заменим полученные блоки
        preparedMessage = message
        for oldString, newString in replaceBlocks:
            preparedMessage = preparedMessage.replace(oldString, newString)

        log.debug("Prepared message: \"%s\"", preparedMessage)
        return preparedMessage

    def __eval(self, expression, check, card, impact):
        return eval(expression)

class EasyStructureImpact(object):

    def __init__(self, posnum, qty, sum):
        self.posnum = posnum
        self.qty = qty
        self.sum = sum

    def __repr__(self):
        return "posnum=%s;qty=%s;sum=%s"%(self.posnum,self.qty,self.sum)

class EasyPureImpact(object):
    pass

class Subsystem(object):
    '''
    classdocs
    '''
    def __init__(self, discountSystem=None):
        if discountSystem != None:
            self.rootCampaignGroup = discountSystem.rootCampaignGroup
            self.discountSystemNumber = discountSystem.number

    def refreshTagsValid (self):
        if self.rootCampaignGroup != None:
            objects = self.rootCampaignGroup.discountCampaigns + self.rootCampaignGroup.childGroups
            for obj in objects:
                obj.tagsValid = None

    def initWithYaml(self, discountSystemYaml):
        self.rootCampaignGroup = yaml.load(discountSystemYaml)

    def getImpacts(self, object):
        """
        Get impacts
        """
        
        if not self.rootCampaignGroup:
            return None

        checkWrapper = CheckWrapper(object)
        getStrCheckPositions(checkWrapper)
 
        return BaseImpacts(checkWrapper, self.rootCampaignGroup, self.discountSystemNumber).getImpacts()

    def removeUnusedProperties(self, impact):
        '''
        Удаление неиспользуемых значений из воздействия
        '''
        for prop in ['campaign', 'objects', 'conditions', 'discountRate', 'clientText', 'cashText', 'active', 'notPosConditionTrue']:
            impact.discount.__dict__.pop(prop, None)

        if impact.discount.counters == []:
            impact.discount.__dict__.pop('counters', None)

        for prop in ['campaign', 'parentGroup', 'discounts', 'parent', 'beginBonusesIntervalCondition',
                     'beginBonusesIntervalValue', 'beginDateWithBeginCurrentMetric', 'beginDateWithBeginNextMetric',
                     'beginBonusesIntervalMetric', 'endBonusesIntervalMetric', 'endBonusesIntervalCondition',
                     'endBonusesIntervalValue', 'beginBonusesDateCalculate', 'endBonusesDateCalculate',
                     'beginRangeDate', 'endRangeDate', 'checkedTags', 'tagsIsAllowed', 'tagsValid', 'beginDate', 'endDate',
                     'clientText', 'cashText', 'conditions', 'minPriceIgnored', 'objects', 'discountRate', 'active',
                     'notPosConditionTrue', 'interactionType', 'allowedTags', 'forbiddenTags', 'discountReverse', 'counters']:

            impact.campaign.__dict__.pop(prop, None)

        for prop in ['checkWrapper', 'items', 'mode']:
            impact.discountRate.__dict__.pop(prop, None)

    # создание структурного воздействия для передачи на Кассу
    def quick(self, check):

        impactGroup = self.getImpacts(check)

        if not impactGroup:
            return [], [], [], [], [], [], []

        screen_message = impactGroup.screenMessage
        receipt_message = impactGroup.receiptMessage

        if log.isEnabledFor(logging.DEBUG):
            log.debug(impactGroup)
            log.debug(u'Сообщения на экран кассира:')
            for str in impactGroup.screenMessage:
                for l in str.split('##'):
                    log.debug(l.encode('utf-8'))
                log.debug(u'-------------')
            log.debug(u'Сообщения для печати в чеке:')
            for str in impactGroup.receiptMessage:
                for l in str.split('##'):
                    log.debug(l.encode('utf-8'))
                log.debug(u'-------------')

        modify={}
        parentSubs={}
        for x in filter(lambda  x:x.type == ImpactType.ADD, impactGroup.getImpacts()):
            log.debug("ADD %d parent %d" % (x.impactObject.positionNumber, x.parentPositionNumber))
            parentPosition = parentSubs.get(x.parentPositionNumber, x.parentPositionNumber)
            parentSubs[x.impactObject.positionNumber]=x.parentPositionNumber
            if parentPosition not in modify:
                modify[parentPosition]=[]
            mod=EasyStructureImpact()
            mod.posnum=x.impactObject.positionNumber
            mod.qty=x.qty
            mod.sum=impactGroup.check.getPositionByNumber(mod.posnum)["sumb"]
            modify[parentPosition].append(mod)

        impacts=[]
        for x in filter(lambda  x:x.type == ImpactType.BRANCH, impactGroup.getImpacts()):
            modify[x.impactObject.positionNumber] = [ EasyStructureImpact(det.positionNumber, det.quantity, det.sumb) for det in x.branchDetails ]

        for x in filter(lambda x:x.type == ImpactType.PURE, impactGroup.getImpacts()):
            impact=EasyPureImpact()
            impact.discountMode = ("%s" % x.discountMode).lower()
            impact.absoluteValue = x.absoluteValue
            impact.positionNumber = x.impactObject.positionNumber
            impact.quantity = float(x.qty) if x.qty else x.qty

            impact.discountRate = copy.copy(x.discountRate if x.discountRate else x.discount.discountRate)

            impact.discount = copy.copy(x.discount)
            impact.campaign = copy.copy(x.discount.campaign)

            impact.cards = [{"codeinternal": x.cards[0]['codeinternal']}] if x.cards else []
            impact.coupons = [{"number": x.coupons[0]['number']}] if x.coupons else []

            impact.campaign.id = x.campaign.id
            impact.campaign.name = x.campaign.name
            if isinstance(impact.campaign, DiscountCampaign):
                impact.campaign.calcAssetDates()
                for c in impactGroup.coupons:
                    if c.campaignId == impact.campaign.id:
                        c.beginDate = impact.campaign.assetBeginDate
                        c.endDate = impact.campaign.assetEndDate

                for c in impactGroup.counters:
                    if c.campaignId == impact.campaign.id:
                        c.beginDate = impact.campaign.assetBeginDate
                        c.endDate = impact.campaign.assetEndDate

            self.removeUnusedProperties(impact)

            impacts.append(impact)

        uniteImpacts = {}
        for x in filter(lambda  x:x.type == ImpactType.UNITE, impactGroup.getImpacts()):
            uniteImpacts[x.impactObject.positionNumber] = [ EasyStructureImpact(pos, quant, None) for pos, quant in x.uniteDetails.items() ]

        coupons_res = []
        for c in impactGroup.coupons:
            coupon = Coupon()
            coupon = copy.deepcopy(c)
            coupons_res.append(coupon)

        counters_res = []
        for c in impactGroup.counters:
            counter = Counter()
            counter = copy.deepcopy(c)
            counters_res.append(counter)

        return impacts, modify, screen_message, receipt_message, uniteImpacts, coupons_res, counters_res

    def encode(self,result):
        return jsonpickle.encode(result, unpicklable=False)

    def quickInJson(self, check):
        # очищаем кэш с резутатами вычислений условий и расчетных ставок
        conditionMemory.clear()
        rateMemory.clear()
        impacts, modify, screen_message, receipt_message, uniteImpacts, coupons, counters = self.quick(check)
        result = [impacts, modify]
        result.append(screen_message)
        result.append(receipt_message)
        result.append(uniteImpacts)
        result.append(coupons)
        result.append(counters)
        return self.encode(result)
